/*
 * SketchyLISP -- An interpreter for purely applicative Scheme
 * Copyright (C) 2005,2006 Nils M Holm <nmh@t3x.org>
 * This is the interpreter shell.
 * See the file LICENSE for conditions of use.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include "sketchy.h"

#define TEXTLEN	SK_TEXTLEN

#define pr(s)		sk_pr(s)
#define nl()		sk_nl()
#define prnum(n,w)	sk_prnum(n,w)

char	Image[TEXTLEN];
int	Nodes, Vcells;
int	PrintVoid;
int	WrongMC;
int	Batch;

void	bye(char *p);
void	catchIntr(int sig);
void	dumpImage(char *p);
char	*findSource(char *p, char *buf);
void	gc_stats(char *p);
int	getName(char *p);
int	getOptNum(int argc, char **argv, int *pi, int *pj, int *pk);
void	getOpts(int argc, char **argv);
void	help(void);
void	help_meta(char *p);
void	help_input(void);
void	init(void);
void	license(char *p);
void	load(char *p);
void	meta(void);
int	meta2(void);
char	*opening(char *p);
void	repl(void);
void	report(char *msg);
void	require(char *p);
int	skip(int all);
void	trace(char *p);
void	usage(void);
void	version(char *p);
int	yesno(char *p, int *pf);

/* Print usage and fail. */
void usage(void) {
	fprintf(stderr,
		"Usage: sketchy [-bi] [-n nodes] [-m vcells] [image]\n");
}

/* Long usage. */
void help(void) {
	fputc('\n', stderr);
	usage();
	fprintf(stderr,
		"\n"
		"-b    batch mode (quiet, exit on first error)\n"
		"-B    batch mode, but print void results\n"
		"-i    init mode (do not load any image)\n"
		"-m #  number of vector cells to allocate (default: %dK)\n"
		"-n #  number of nodes to allocate (default: %dK)\n"
		"\n"
		"default image: %s\n\n",
		SK_DFL_NODES/1024, SK_DFL_VCELLS/1024, SK_DFL_IMAGE);
}

/*
 * Get numeric option argument.
 * pi,pj,pk are pointers to i,j,k of getOpts().
 */
int getOptNum(int argc, char **argv, int *pi, int *pj, int *pk) {
	int	n, c;

	if (++(*pi) >= argc) {
		usage();
		exit(1);
	}
	n = atoi(argv[*pi]);
	c = argv[*pi][strlen(argv[*pi])-1];
	switch (c) {
	case 'K':	n = n * 1024; break;
	case 'M':	n = n * 1024 * 1024; break;
	}
	*pj = *pk = 0;
	return n;
}

/* Set defaults and evaluate command line options. */
void getOpts(int argc, char **argv) {
	char	*a;
	int	i, j, k;

	memset(Image, 0, TEXTLEN);
	k = strlen(SK_DFL_IMAGE);
	strncpy(Image, SK_DFL_IMAGE, k>=TEXTLEN-1? TEXTLEN: k);
	Nodes = SK_DFL_NODES;
	Vcells = SK_DFL_VCELLS;
	PrintVoid = 1;
	Batch = 0;
	i = 1;
	while (i < argc) {
		a = argv[i];
		if (a[0] != '-') break;
		k = strlen(a);
		for (j=1; j<k; j++) {
			switch (a[j]) {
			case 'b':
				Batch = 1;
				PrintVoid = 0;
				break;
			case 'B':
				Batch = 1;
				PrintVoid = 1;
				break;
			case 'n':
				Nodes = getOptNum(argc, argv, &i, &j, &k);
				break;
			case 'm':
				Vcells = getOptNum(argc, argv, &i, &j, &k);
				break;
			case 'i':
				Image[0] = 0;
				break;
			case '?':
			case 'h':
				help();
				exit(1);
				break;
			default:
				usage();
				exit(1);
			}
		}
		i = i+1;
	}
	if (i < argc) {
		memset(Image, 0, TEXTLEN);
		k = strlen(a);
		strncpy(Image, a, k>=TEXTLEN-1? TEXTLEN: k);
	}
	if (Nodes < SK_MIN_SIZE || Vcells < SK_MIN_SIZE) {
		fprintf(stderr, "sketchy: minimum pool size is %d\n",
			SK_MIN_SIZE);
		exit(1);
	}
}

/* Stop interpreter when catching SIGINT. */
void catchIntr(int sig) {
	sig = 0; /*LINT*/
	sk_stop();
	signal(SIGINT, catchIntr);
}

/* Report meta command error. */
void report(char *msg) {
	printf("Meta command error: %s\n", msg);
	WrongMC = 1;
}

/* Report full path of source file. */
char *opening(char *p) {
	if (!Batch) printf("loading %s\n", p);
	return p;
}

/*
 * Locate source file.
 * Try: ./file
 *      ./file.scm (if no .scm suffix is present)
 *      P/.file
 *      P/.file.scm (where P is a path from $SKETHYSRC).
 * SKETCHYSRC is a colon-separated list of paths.
 */
char *findSource(char *p, char *buf) {
	char	*sksrc, *q;
	char	*toolong;
	FILE	*f;
	int	k;

	toolong = "path too long in :load or :require";
	if (*p == '.' || *p == '/' || *p == '~') return opening(p);
	sksrc = getenv("SKETCHYSRC");
	if (sksrc == NULL) {
		puts("! In case source files cannot be located, try");
		puts("! setting the SKETCHYSRC environment variable.");
	}
	if ((f = fopen(p, "r")) != NULL) {
		fclose(f);
		return opening(p);
	}
	k = strlen(p);
	if (k < 4 || strcmp(&p[k-4], ".scm")) {
		if (strlen(p) + 4 >= SK_MAXPATHL-1) {
			report(toolong);
			return NULL;
		}
		strcpy(buf, p);
		strcat(buf, ".scm");
		f = fopen(buf, "r");
		if (f) return opening(buf);
	}
	while (sksrc && *sksrc) {
		q = strchr(sksrc, ':');
		if (q) *q = 0;
		if (strlen(sksrc) + strlen(p) >= SK_MAXPATHL-2) {
			report(toolong);
			return NULL;
		}
		sprintf(buf, "%s/%s", sksrc, p);
		if (q) *q = ':';
		f = fopen(buf, "r");
		if (f) return opening(buf);
		k = strlen(buf);
		if (k < 4 || strcmp(&buf[k-4], ".scm")) {
			if (strlen(buf) + 4 >= SK_MAXPATHL-1) {
				report(toolong);
				return NULL;
			}
			strcat(buf, ".scm");
			f = fopen(buf, "r");
			if (f) return opening(buf);
		}
		sksrc = q? &q[1]: "";
	}
	report("could not locate input file.");
	return NULL;
}

/* Load expressions from file. */
void load(char *p) {
	char	path[SK_MAXPATHL];

	p = findSource(p, path);
	if (p == NULL) return;
	sk_load(p);
	if (!Batch && !SK_errFlag) {
		pr("loaded"); nl();
	}
}

/* Load expressions if not already present. */
void require(char *p) {
	int	r;
	char	path[SK_MAXPATHL];

	p = findSource(p, path);
	if (p == NULL) return;
	r = sk_require(p);
	if (!Batch && !SK_errFlag) {
		pr(r? "loaded": "cached"); nl();
	}
}

/*
 * Read characters from the input device until a non-space
 * character [^\t ] is found. Return that character.
 * If all=1, spaces = [ \t\n\r].
 */
int skip(int all) {
	int	c;
	char	*set;

	c = sk_rdch();
	set = all? " \t\n\r": " \t";
	while (strchr(set, c)) c = sk_rdch();
	return c;
}

/*
 * Read all characters up to the next delimiter (white space or EOT)
 * from the input device and store them in the vector P.
 * Convert each character read to lower case.
 * Return the terminating character.
 * This routine is used to accept parameters to meta commands.
 */
int getName(char *p) {
	int	c, i;

	c = skip(0);
	i = 0;
	while (	c != ' ' && c != '\t' && c != '\n' && c != '\r' &&
		c != SK_EOT
	) {
		p[i] = c;
		if (i >= TEXTLEN-2) {
			report("argument too long");
			return 0;
		}
		c = sk_rdch();
		i = i+1;
	}
	p[i] = 0;
	return c;
}

/*
 * Extract a YES/NO, ON/OFF, 0/1 argument used to
 * enable and disable options.
 */
int yesno(char *p, int *pf) {
	int	c, f;

	c = getName(p);
	f = 2;
	if (p[0] == 'y' || p[0] == '1') f = -1;
	if (p[0] == 'n' || p[0] == '0') f = 0;
	if (p[0] == 'o') {
		if (p[1] == 'n') f = -1;
		if (p[1] == 'f') f = 0;
	}
	if (f == 2) report("commands expects on/off argument");
	pf[0] = f;
	return c;
}

/* Print conditions of use. */
void license(char *p) {
	char	**lp;

	p = NULL;
	lp = sk_license();
	while (*lp) {
		pr(*lp); nl();
		lp++;
	}
}

/* Normal termination. */
void bye(char *p) {
	p = NULL;
	sk_bye();
}

/* Dump workspace to file. */
void dumpImage(char *p) {
	sk_dumpImage(p);
	if (!Batch && !SK_errFlag) {
		pr("image dumped"); nl();
	}
}

/* Display help on entering meta commands. */
void help_input(void) {
	pr("Meta commands may be abbreviated as long as an"); nl();
	pr("abbreviation is unambiguous:"); nl();
	pr(":a 1  is the same as  :arrow-mode 1"); nl();
	pr("but  :show  does not work, because it may expand"); nl();
	pr("to either  :show-license  or  :show-version."); nl();
	nl();
	pr("Commands containing a dash may be abbreviated by"); nl();
	pr("using the first character of the command plus the"); nl();
	pr("character after the dash:"); nl();
	pr(":sv  is the same as  :show-version"); nl();
	nl();
	pr("Flags turn switches on or off. The values 'yes',"); nl();
	pr("'on', and '1' activate a switch, and the values"); nl();
	pr("'no', 'off', and '0' deactivate it."); nl();
}

/* Display meta command summary. */
void help_meta(char *p) {
	if (p[0] && !strncmp(p, "input", strlen(p))) {
		help_input();
		return;
	}
	pr(":arrow-mode 0..2   - 0=symbol, 1=comment, 2=verify"); nl();
	pr(":closure-form 0..2 - set closure print level"); nl();
	pr(":dump-image file   - dump image to file"); nl();
	pr(":dump-sym [sym]    - dump symbol table"); nl();
	pr(":gc                - collect garbage"); nl();
	pr(":help [input]      - print help text (also ?)"); nl();
	pr(":load file         - load file (~/file = $HOME/file"); nl();
	pr("                     and =file = $SKETCHYSRC/file)"); nl();
	pr(":quit              - quit"); nl();
	pr(":r5rs-apply flag   - restricted (R5RS) APPLY on/off"); nl();
	pr(":require file      - load conditionally (see :load)"); nl();
	pr(":show-license      - print conditions of use"); nl();
	pr(":show-version      - print version"); nl();
	pr(":statistics flag   - statistics on/off"); nl();
	pr(":trace [name]      - trace function (:trace = off)"); nl();
}

/* Do garbage collection and display some statistics. */
void gc_stats(char *p) {
	struct sk_gcStats	gs;

	p = NULL; /*LINT*/
	sk_gc(&gs);
	prnum(gs.nodes_max, 7);
	pr(" nodes used (peak)"); nl();
	prnum(gs.nodes_used, 7);
	pr(" nodes in use"); nl();
	prnum(gs.vcells_max, 7);
	pr(" vector cells used (peak)"); nl();
	prnum(gs.vcells_used, 7);
	pr(" vector cells in use"); nl();
}

/* Register function to trace. */
void trace(char *p) {
	SK_trace = sk_findSym(p);
	if (SK_trace != sk_nil()) {
		if (!Batch) {
			pr("tracing "); pr(p); nl();
		}
	}
	else if (p[0]) {
		report("no such function");
	}
	else {
		if (!Batch) {
			pr("trace off"); nl();
		}
	}
}

/* Print version summary. */
void version(char *p) {
	pr("SketchyLISP (C) 2006 Nils M Holm"); nl();
	pr("Latest modification: ");
	pr(SK_RELEASE); nl();
	prnum(Nodes, 0);
	pr(" nodes, ");
	prnum(Vcells, 0);
	pr(" vector cells"); nl();
}

#define MT_FLAG	0	/* command has a boolean argument */
#define MT_INT	1	/* command has an integer argument */
#define MT_PROC	2	/* command handles args on its own */

struct MetaCmd {
	char	*cmd;
	int	type;
	int	lo, hi;
	int	*pval;
	void	(*fn)(char *s);
} MetaCmds[] = {
	{	"arrow-mode",	MT_INT, 0,2,	&SK_arrowMode,	NULL	},
	{	"closure-form",	MT_INT, 0,2,	&SK_closureForm,NULL	},
	{	"dump-image",	MT_PROC, 1,0,	NULL,	dumpImage	},
	{	"dump-symbols",	MT_PROC, 0,0,	NULL,	sk_dumpSymbols	},
	{	"gc",		MT_PROC, 0,0,	NULL,	gc_stats	},
	{	"help",		MT_PROC, 0,0,	NULL,	help_meta	},
	{	"?",		MT_PROC, 0,0,	NULL,	help_meta	},
	{	"load",		MT_PROC, 1,0,	NULL,	load		},
	{	"quit",		MT_PROC, 0,0,	NULL,	bye		},
	{	"r5rs-apply",	MT_FLAG, 0,0,	&SK_strictApply,NULL	},
	{	"require",	MT_PROC, 1,0,	NULL,	require		},
	{	"show-license",	MT_PROC, 0,0,	NULL,	license		},
	{	"show-version",	MT_PROC, 0,0,	NULL,	version		},
	{	"statistics",	MT_FLAG, 0,0,	&SK_statFlag,	NULL	},
	{	"trace",	MT_PROC, 0,0,	NULL,	trace		},
	{	NULL,		0, 0,0,		NULL,		NULL	},
};

/* Read and process meta (colon) commands. */
int meta2(void) {
	static char	s[TEXTLEN];
	int		i, k;
	char		*p, *q;
	int		v;
	int		c;
	
	WrongMC = 1;
	c = getName(s);
	k = strlen(s);
	for (i=0; MetaCmds[i].cmd; i++) {
		if (!strncmp(MetaCmds[i].cmd, s, k)) {
			if (	MetaCmds[i+1].cmd &&
				!strncmp(MetaCmds[i+1].cmd, s, k)
			)
				report("ambiguous meta command (:h = help)");
			else
				WrongMC = 0;
			break;
		}
	}
	if (WrongMC && !MetaCmds[i].cmd && strlen(s) == 2) {
		/* Try abbreviations */
		for (i=0; MetaCmds[i].cmd; i++) {
			p = MetaCmds[i].cmd;
			q = strchr(p, '-');
			if (q != NULL && p[0] == s[0] && q[1] == s[1]) {
				WrongMC = 0;
				break;
			}

		}
	}
	if (!MetaCmds[i].cmd) {
		report("unknown meta command (:h = help)");
		return c;
	}
	s[0] = 0;
	if (c == '\n' || c == SK_EOT) {
		/* no argument */
	}
	else if (MetaCmds[i].type == MT_FLAG) {
		c = yesno(s, &v);
	}
	else if (MetaCmds[i].type == MT_INT) {
		c = getName(s);
		v = atoi(s);
		if (v < MetaCmds[i].lo || v > MetaCmds[i].hi)
			report("argument out of range");
	}
	else {
		c = getName(s);
	}
	if (!WrongMC && !s[0] && MetaCmds[i].type != MT_PROC)
		report("missing argument");
	if (WrongMC) return c;
	if (MetaCmds[i].type == MT_FLAG || MetaCmds[i].type == MT_INT) {
		*(MetaCmds[i].pval) = v;
		if (!Batch) {
			pr(MetaCmds[i].cmd);
			pr(" = ");
			if (MetaCmds[i].type == MT_FLAG)
				pr(v? "on": "off");
			else
				prnum(v, 0);
			nl();
		}
	}
	else {
		if (MetaCmds[i].lo && !s[0])
			report("missing argument");
		else
			(*MetaCmds[i].fn)(s);
	}
	return c;
}

/* Run meta command interpreter and clean up. */
void meta(void) {
	int	c;

	c = meta2();
	while (c != '\n' && c != SK_EOT) c = sk_rdch();
	if (SK_errFlag) sk_printError();
}

/* The Read-Eval-Print Loop (REPL). */
void repl(void) {
	int	n, c;

	while(1) {
		SK_errFlag = 0;
		c = skip(1);
		if (c == SK_metaChar) {
			meta();
			continue;
		}
		sk_reject(c);
		n = sk_read();
		if (n == sk_eof()) return;
		if (SK_errFlag) {
			sk_printError();
			if (Batch) exit(1);
			continue;
		}
		n = sk_eval(n);
		if (SK_errFlag) {
			sk_printError();
			if (Batch) exit(1);
		}
		else {
			if (PrintVoid || n != sk_void()) {
				if (!Batch) pr("=> ");
				sk_print(n); sk_nl();
			}
			if (SK_statFlag) sk_printStats();
		}
	}
}

/* Launch interpreter. */
void init(void) {
	if (sk_init(Nodes, Vcells)) {
		fprintf(stderr, "sketchy init failed (memory problem)\n");
		exit(1);
	}
}

/*
 * Here the fun begins.
 */
int main(int argc, char **argv) {
	getOpts(argc, argv);
	init();
	getOpts(argc, argv);
	if (!Batch) {
		pr("SketchyLISP ");
		pr(SK_RELEASE);
		pr(" (C) 2006 Nils M Holm"); sk_nl();
	}
	if (Image[0]) {
		if (sk_loadImage(Image)) {
			sk_printError();
			if (Batch) exit(1);
			sk_fini();
			init();
			getOpts(argc, argv);
		}
	}
	else if (!Batch) {
		pr("Warning: no image loaded"); sk_nl();
	}
	signal(SIGINT, catchIntr);
	repl();
	sk_bye();
	return 0; /*LINT*/
}
