; SketchyLISP -- An interpreter for purely applicative Scheme
; Copyright (C) 2006 Nils M Holm <nmh@t3x.org>
; See the file LICENSE for conditions of use.

(package 'core)

; Syntax transformer

; Extend environment
(define (ext-env x v env)
  (cons (cons x v) env))

; Match form against pattern.
; name - name of syntax transformer;
; lits - literals (keywords) of syntax-rules;
; env - initial environment.
; Returns an environment with bindings of
; pattern variables to subforms or #F in
; cass form does not match pattern.

(define (match form pattern name lits env)
  (letrec
    ((_match (lambda (form pattern env)
      (cond ((eq? pattern '_)
          (if (eq? form name) env #f))
        ((equal? pattern '(...))
          (ext-env '... form env))
        ((memq pattern lits)
          (if (eq? form pattern) env #f))
        ((symbol? pattern)
          (if (memq pattern lits)
              (if (eq? pattern form) env #f)
              (ext-env pattern form env)))
        ((and (pair? pattern)
              (pair? form))
          (let ((e (_match (car form) (car pattern) env)))
            (and e (_match (cdr form) (cdr pattern) e))))
        (#t (if (equal? form pattern) env #f))))))
    (_match form pattern env)))

; Find rule with a pattern that matches the given form.
; rules - list of rules to try;
; name - name of syntax transformer;
; lits - literals (keywords) of syntax-rules.

(define (find-rule form rules name lits)
  (cond ((null? rules)
      (bottom name ': 'bad 'syntax))
    (#t (let ((e (match form (caar rules) name lits '())))
          (cond (e (list (cadar rules) e))
            (#t (find-rule form (cdr rules) name lits)))))))

; Substitute keys of env by values of env in form.
(define (substitute form env)
  (cond ((equal? form '(...))
      (let ((v (assq '... env)))
        (if v (cdr v)
              (bottom 'unmatched '... 'in 'syntax-rule))))
    ((symbol? form)
      (let ((v (assq form env)))
        (if v (cdr v) form)))
    ((pair? form)
      (cons (substitute (car form) env)
            (substitute (cdr form) env)))
    (#t form)))

; Transform the syntax of the given form.
; This is the entry point of the syntax transformer
; when called from the SketchyLISP core.

(define (transform-syntax form)
  (let ((syn (syntax-of (car form))))
    (if (not syn)
        (bottom 'bad 'syntax 'rules: syntax)
        (let ((name (car syn))
              (lits (cadr syn))
              (rules (caddr syn)))
          (let ((rule+env (find-rule form rules name lits)))
            (transform-all
              (substitute (car rule+env)
                          (cadr rule+env))))))))

; TRANSFORM-ALL *must* be defined *after* TRANSFORM-SYNTAX.
; Probably because TRANSFORM-SYNTAX is defined by the
; SketchyLISP core. To be investigated!

; Syntax-transform all subforms of the given form.
(define (transform-all form)
  (cond ((not (pair? form)) form)
    ((eq? (car form) 'quote) form)
    ((syntax-of (car form))
      (transform-syntax form))
    (#t (map transform-all form))))

(package)

